@@ -38,11 +38,12 @@
 #include <math.h>
 
 
-#define TIMEOUT_MS  (1000U)
-#define SENSOR_COMMAND_TIMEOUT_MS  (3000U)
-#define FRAME_SIZE_BYTES           (512U)
+#define TIMEOUT_MS                  (1000U)
+#define SENSOR_COMMAND_TIMEOUT_MS   (3000U)
+#define FRAME_SIZE_BYTES            (512U)
 #define ROE JSDRV_RETURN_ON_ERROR
-#define SAMPLING_FREQUENCY (2000000U)
+#define SAMPLING_FREQUENCY          (2000000U)
+#define STREAM_PAYLOAD_FULL         (JSDRV_STREAM_DATA_SIZE - JSDRV_STREAM_HEADER_SIZE - JS220_USB_FRAME_LENGTH)
 
 struct js110_dev_s;  // forward declaration, see below
 
@@ -282,7 +283,7 @@ static const struct param_s PARAMS[] = {
         on_i_range_post,
     },
     {
-        "s/fs",
+        "h/fs",
         "{"
             "\"dtype\": \"u32\","
             "\"brief\": \"The sampling frequency.\","
@@ -424,7 +425,6 @@ const struct field_def_s FIELDS[] = {
 };
 
 struct port_s {
-    uint64_t sample_id_next;
     struct jsdrvp_msg_s * msg;
     struct jsdrv_downsample_s * downsample;
 };
@@ -439,6 +439,7 @@ struct js110_dev_s {
     uint64_t packet_index;
     struct js110_sp_s sample_processor;
     struct js110_stats_s stats;
+    uint64_t sample_id;
 
     struct port_s ports[JSDRV_ARRAY_SIZE(FIELDS)];
 
@@ -774,18 +775,22 @@ static void on_i_range_mode(struct js110_dev_s * d, const struct jsdrv_union_s *
 }
 
 static void on_i_range_pre(struct js110_dev_s * d, const struct jsdrv_union_s * value) {
+    JSDRV_LOGI("on_i_range_pre %d", (int) value->value.u8);
     d->sample_processor._suppress_samples_pre = value->value.u8;
 }
 
 static void on_i_range_win(struct js110_dev_s * d, const struct jsdrv_union_s * value) {
+    JSDRV_LOGI("on_i_range_win %d", (int) value->value.u8);
     js110_sp_suppress_win(&d->sample_processor, value->value.u8);
 }
 
 static void on_i_range_win_sz(struct js110_dev_s * d, const struct jsdrv_union_s * value) {
+    JSDRV_LOGI("on_i_range_win_sz %d", (int) value->value.u8);
     d->sample_processor._suppress_samples_window = value->value.u8;
 }
 
 static void on_i_range_post(struct js110_dev_s * d, const struct jsdrv_union_s * value) {
+    JSDRV_LOGI("on_i_range_post %d", (int) value->value.u8);
     d->sample_processor._suppress_samples_post = value->value.u8;
 }
 
@@ -797,13 +802,17 @@ static void on_sampling_frequency(struct js110_dev_s * d, const struct jsdrv_uni
         return;
     }
     fs = v.value.u32;
-    JSDRV_LOGW("on_sampling_frequency(%lu)", fs);
+    JSDRV_LOGI("on_sampling_frequency(%lu)", fs);
     for (uint32_t idx = 0; idx < JSDRV_ARRAY_SIZE(FIELDS); ++idx) {
         struct port_s *p = &d->ports[idx];
         if (p->downsample) {
             jsdrv_downsample_free(p->downsample);
             p->downsample = NULL;
         }
+        if (p->msg) {
+            jsdrvp_msg_free(d->context, p->msg);
+            p->msg = NULL;
+        }
         p->downsample = jsdrv_downsample_alloc(SAMPLING_FREQUENCY, fs);
         if (NULL == p->downsample) {
             JSDRV_LOGW("jsdrv_downsample_alloc failed");
@@ -825,11 +834,14 @@ static void on_update_ctrl(struct js110_dev_s * d, const struct jsdrv_union_s *
                 struct port_s * p = &d->ports[idx];
                 struct jsdrvp_msg_s *m = p->msg;
                 p->msg = NULL;
-                p->sample_id_next = 0;
                 if (NULL != m) {
                     jsdrvp_msg_free(d->context, m);
                 }
+                if (NULL != p->downsample) {
+                    jsdrv_downsample_clear(p->downsample);
+                }
             }
+            d->sample_id = 0;
         }
         stream_settings_send(d);
         JSDRV_LOGI("on_update_ctrl %d (stream change complete)", param);
@@ -939,6 +951,13 @@ static int32_t d_close(struct js110_dev_s * d) {
     struct jsdrvp_msg_s * m = jsdrvp_msg_alloc_value(d->context, JSDRV_MSG_CLOSE, &jsdrv_union_u32(0));
     msg_queue_push(d->ll.cmd_q, m);
     m = ll_await_topic(d, JSDRV_MSG_CLOSE, TIMEOUT_MS);
+    for (uint32_t idx = 0; idx < JSDRV_ARRAY_SIZE(d->ports); ++idx) {
+        struct port_s * p = &d->ports[idx];
+        if (NULL != p->msg) {
+            jsdrvp_msg_free(d->context, p->msg);
+            p->msg = NULL;
+        }
+    }
     if (!m) {
         JSDRV_LOGW("ll_driver open timed out");
         return JSDRV_ERROR_TIMED_OUT;
@@ -1029,13 +1048,15 @@ static struct jsdrvp_msg_s * field_message_get(struct js110_dev_s * d, uint8_t f
         m = jsdrvp_msg_alloc_data(d->context, "");
         tfp_snprintf(m->topic, sizeof(m->topic), "%s/%s", d->ll.prefix, field_def->data_topic);
         s = (struct jsdrv_stream_signal_s *) m->value.value.bin;
-        s->sample_id = p->sample_id_next;
+        s->sample_id = d->sample_id;
         s->index = field_def->index;
         s->field_id = field_def->field_id;
         s->element_type = field_def->element_type;
         s->element_size_bits = field_def->element_size_bits;
+        s->sample_rate = SAMPLING_FREQUENCY;
+        s->decimate_factor = jsdrv_downsample_decimate_factor(p->downsample);
         s->element_count = 0;
-        m->u32_a = (uint32_t) p->sample_id_next;
+        m->u32_a = (uint32_t) d->sample_id;
         m->value.app = JSDRV_PAYLOAD_TYPE_STREAM;
         m->value.size = JSDRV_STREAM_HEADER_SIZE;
         p->msg = m;
@@ -1049,12 +1070,11 @@ static void field_message_process_end(struct js110_dev_s * d, uint8_t idx) {
     struct jsdrvp_msg_s * m = p->msg;
     struct jsdrv_stream_signal_s * s = (struct jsdrv_stream_signal_s *) m->value.value.bin;
     ++s->element_count;
-    ++p->sample_id_next;
     if ((s->element_size_bits < 8) && (((s->element_count * s->element_size_bits) & 0x7) != 0)) {
         return;
     }
     uint32_t element_count_max = SAMPLING_FREQUENCY / (20 * jsdrv_downsample_decimate_factor(p->downsample));
-    if ((((s->element_count * s->element_size_bits) / 8) >= JSDRV_STREAM_DATA_SIZE)
+    if ((((s->element_count * s->element_size_bits) / 8) >= STREAM_PAYLOAD_FULL)
             || (s->element_count >= element_count_max)) {
         jsdrvp_backend_send(d->context, p->msg);
         p->msg = NULL;
@@ -1133,6 +1153,7 @@ static void handle_sample(struct js110_dev_s * d, uint32_t sample, uint8_t v_ran
     add_u4_field(d, 3, z.current_range);
     add_u1_field(d, 4, z.gpi0);
     add_u1_field(d, 5, z.gpi1);
+    ++d->sample_id;
 
     struct jsdrv_statistics_s * s = js110_stats_compute(&d->stats, z.i, z.v, z.p);
     if (NULL != s) {
@@ -1143,6 +1164,7 @@ static void handle_sample(struct js110_dev_s * d, uint32_t sample, uint8_t v_ran
         m->value = jsdrv_union_cbin_r((uint8_t *) dst, sizeof(*dst));
         m->value.app = JSDRV_PAYLOAD_TYPE_STATISTICS;
         jsdrvp_backend_send(d->context, m);
+        s->block_sample_id = d->sample_id;
     }
 }
 
@@ -1247,7 +1269,6 @@ static THREAD_RETURN_TYPE driver_thread(THREAD_ARG_TYPE lpParam) {
             ;
         }
     }
-
     JSDRV_LOGI("JS110 USB upper-level thread done %s", d->ll.prefix);
     THREAD_RETURN();
 }
@@ -1257,6 +1278,18 @@ static void join(struct jsdrvp_ul_device_s * device) {
     jsdrvp_send_finalize_msg(d->context, d->ul.cmd_q, "");
     // and wait for thread to exit.
     jsdrv_thread_join(&d->thread, 1000);
+
+    for (uint32_t idx = 0; idx < JSDRV_ARRAY_SIZE(d->ports); ++idx) {
+        struct port_s *p = &d->ports[idx];
+        if (NULL != p->msg) {
+            jsdrvp_msg_free(d->context, p->msg);
+            p->msg = NULL;
+        }
+        if (NULL != p->downsample) {
+            jsdrv_downsample_free(p->downsample);
+            p->downsample = NULL;
+        }
+    }
     jsdrv_free(d);
 }
 
@@ -1273,6 +1306,7 @@ int32_t jsdrvp_ul_js110_usb_factory(struct jsdrvp_ul_device_s ** device, struct
     d->state = ST_CLOSED;
     on_sampling_frequency(d, &jsdrv_union_u32(SAMPLING_FREQUENCY));
     js110_sp_initialize(&d->sample_processor);
+    js110_stats_initialize(&d->stats);
 
     for (int i = 0; NULL != PARAMS[i].topic; ++i) {
         jsdrv_meta_default(PARAMS[i].meta, &d->param_values[i]);
